package javax.swing.tree;

import java.util.EmptyStackException;
import java.util.Stack;
import java.util.Collections;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.NoSuchElementException;
import java.beans.Transient;
import java.util.Vector;
import java.util.Enumeration;
import java.io.Serializable;

public class DefaultMutableTreeNode implements Cloneable, MutableTreeNode, Serializable
{
    private static final long serialVersionUID = -4298474751201349152L;
    public static final Enumeration<TreeNode> EMPTY_ENUMERATION;
    protected MutableTreeNode parent;
    protected Vector children;
    protected transient Object userObject;
    protected boolean allowsChildren;
    
    public DefaultMutableTreeNode() {
        this(null);
    }
    
    public DefaultMutableTreeNode(Object o) {
        this(o, true);
    }
    
    public DefaultMutableTreeNode(Object userObject, boolean allowsChildren) {
        this.parent = null;
        this.allowsChildren = allowsChildren;
        this.userObject = userObject;
    }
    
    @Override
    public void insert(MutableTreeNode obj, int index) {
        if (!this.allowsChildren) {
            throw new IllegalStateException("node does not allow children");
        }
        if (obj == null) {
            throw new IllegalArgumentException("new child is null");
        }
        if (this.isNodeAncestor(obj)) {
            throw new IllegalArgumentException("new child is an ancestor");
        }
        MutableTreeNode mutableTreeNode = (MutableTreeNode)obj.getParent();
        if (mutableTreeNode != null) {
            mutableTreeNode.remove(obj);
        }
        obj.setParent(this);
        if (this.children == null) {
            this.children = new Vector();
        }
        this.children.insertElementAt(obj, index);
        return;
    }
    
    @Override
    public void remove(int index) {
        MutableTreeNode mutableTreeNode = (MutableTreeNode)this.getChildAt(index);
        this.children.removeElementAt(index);
        mutableTreeNode.setParent(null);
    }
    
    @Transient
    @Override
    public void setParent(MutableTreeNode parent) {
        this.parent = parent;
    }
    
    @Override
    public TreeNode getParent() {
        return this.parent;
    }
    
    @Override
    public TreeNode getChildAt(int index) {
        if (this.children == null) {
            throw new ArrayIndexOutOfBoundsException("node has no children");
        }
        return (TreeNode)this.children.elementAt(index);
    }
    
    @Override
    public int getChildCount() {
        if (this.children == null) {
            return 0;
        }
        return this.children.size();
    }
    
    @Override
    public int getIndex(TreeNode o) {
        if (o == null) {
            throw new IllegalArgumentException("argument is null");
        }
        if (!this.isNodeChild(o)) {
            return -1;
        }
        return this.children.indexOf(o);
    }
    
    @Override
    public Enumeration children() {
        if (this.children == null) {
            return DefaultMutableTreeNode.EMPTY_ENUMERATION;
        }
        return this.children.elements();
    }
    
    public void setAllowsChildren(boolean allowsChildren) {
        if (allowsChildren != this.allowsChildren && !(this.allowsChildren = allowsChildren)) {
            this.removeAllChildren();
        }
    }
    
    @Override
    public boolean getAllowsChildren() {
        return this.allowsChildren;
    }
    
    @Override
    public void setUserObject(Object userObject) {
        this.userObject = userObject;
    }
    
    public Object getUserObject() {
        return this.userObject;
    }
    
    @Override
    public void removeFromParent() {
        MutableTreeNode mutableTreeNode = (MutableTreeNode)this.getParent();
        if (mutableTreeNode != null) {
            mutableTreeNode.remove(this);
        }
    }
    
    @Override
    public void remove(MutableTreeNode mutableTreeNode) {
        if (mutableTreeNode == null) {
            throw new IllegalArgumentException("argument is null");
        }
        if (!this.isNodeChild(mutableTreeNode)) {
            throw new IllegalArgumentException("argument is not a child");
        }
        this.remove(this.getIndex(mutableTreeNode));
        return;
    }
    
    public void removeAllChildren() {
        for (int i = this.getChildCount() - 1; i >= 0; --i) {
            this.remove(i);
        }
    }
    
    public void add(MutableTreeNode mutableTreeNode) {
        if (mutableTreeNode != null && mutableTreeNode.getParent() == this) {
            this.insert(mutableTreeNode, this.getChildCount() - 1);
        }
        else {
            this.insert(mutableTreeNode, this.getChildCount());
        }
    }
    
    public boolean isNodeAncestor(TreeNode treeNode) {
        if (treeNode == null) {
            return false;
        }
        while (this != treeNode) {
            if ((this = this.getParent()) != null) {
                continue;
            }
            return false;
        }
        return true;
    }
    
    public boolean isNodeDescendant(DefaultMutableTreeNode defaultMutableTreeNode) {
        return defaultMutableTreeNode != null && defaultMutableTreeNode.isNodeAncestor(this);
    }
    
    public TreeNode getSharedAncestor(DefaultMutableTreeNode treeNode) {
        if (treeNode == this) {
            return this;
        }
        if (treeNode == null) {
            return null;
        }
        int level = this.getLevel();
        int level2 = treeNode.getLevel();
        int i;
        TreeNode treeNode2;
        if (level2 > level) {
            i = level2 - level;
            treeNode2 = treeNode;
        }
        else {
            i = level - level2;
            treeNode2 = this;
            this = treeNode;
        }
        while (i > 0) {
            treeNode2 = treeNode2.getParent();
            --i;
        }
        while (treeNode2 != this) {
            treeNode2 = treeNode2.getParent();
            this = this.getParent();
            if (treeNode2 != null) {
                continue;
            }
            if (treeNode2 != null || this != null) {
                throw new Error("nodes should be null");
            }
            return null;
        }
        return treeNode2;
    }
    
    public boolean isNodeRelated(DefaultMutableTreeNode defaultMutableTreeNode) {
        return defaultMutableTreeNode != null && this.getRoot() == defaultMutableTreeNode.getRoot();
    }
    
    public int getDepth() {
        Object nextElement = null;
        Enumeration<Object> breadthFirstEnumeration = this.breadthFirstEnumeration();
        while (breadthFirstEnumeration.hasMoreElements()) {
            nextElement = breadthFirstEnumeration.nextElement();
        }
        if (nextElement == null) {
            throw new Error("nodes should be null");
        }
        return ((DefaultMutableTreeNode)nextElement).getLevel() - this.getLevel();
    }
    
    public int getLevel() {
        int n = 0;
        while ((this = this.getParent()) != null) {
            ++n;
        }
        return n;
    }
    
    public TreeNode[] getPath() {
        return this.getPathToRoot(this, 0);
    }
    
    protected TreeNode[] getPathToRoot(TreeNode treeNode, int n) {
        TreeNode[] pathToRoot;
        if (treeNode == null) {
            if (n == 0) {
                return null;
            }
            pathToRoot = new TreeNode[n];
        }
        else {
            ++n;
            pathToRoot = this.getPathToRoot(treeNode.getParent(), n);
            pathToRoot[pathToRoot.length - n] = treeNode;
        }
        return pathToRoot;
    }
    
    public Object[] getUserObjectPath() {
        TreeNode[] path = this.getPath();
        Object[] array = new Object[path.length];
        for (int i = 0; i < path.length; ++i) {
            array[i] = ((DefaultMutableTreeNode)path[i]).getUserObject();
        }
        return array;
    }
    
    public TreeNode getRoot() {
        do {
            this = this.getParent();
        } while (this != null);
        return this;
    }
    
    public boolean isRoot() {
        return this.getParent() == null;
    }
    
    public DefaultMutableTreeNode getNextNode() {
        if (this.getChildCount() == 0) {
            DefaultMutableTreeNode nextSibling = this.getNextSibling();
            if (nextSibling == null) {
                DefaultMutableTreeNode defaultMutableTreeNode = (DefaultMutableTreeNode)this.getParent();
                while (defaultMutableTreeNode != null) {
                    DefaultMutableTreeNode nextSibling2 = defaultMutableTreeNode.getNextSibling();
                    if (nextSibling2 != null) {
                        return nextSibling2;
                    }
                    defaultMutableTreeNode = (DefaultMutableTreeNode)defaultMutableTreeNode.getParent();
                }
                return null;
            }
            return nextSibling;
        }
        else {
            return (DefaultMutableTreeNode)this.getChildAt(0);
        }
    }
    
    public DefaultMutableTreeNode getPreviousNode() {
        DefaultMutableTreeNode defaultMutableTreeNode = (DefaultMutableTreeNode)this.getParent();
        if (defaultMutableTreeNode == null) {
            return null;
        }
        DefaultMutableTreeNode previousSibling = this.getPreviousSibling();
        if (previousSibling != null) {
            if (previousSibling.getChildCount() == 0) {
                return previousSibling;
            }
            return previousSibling.getLastLeaf();
        }
        else {
            return defaultMutableTreeNode;
        }
    }
    
    public Enumeration preorderEnumeration() {
        return new DefaultMutableTreeNode.PreorderEnumeration(this);
    }
    
    public Enumeration postorderEnumeration() {
        return new DefaultMutableTreeNode.PostorderEnumeration(this);
    }
    
    public Enumeration breadthFirstEnumeration() {
        return new DefaultMutableTreeNode.BreadthFirstEnumeration(this);
    }
    
    public Enumeration depthFirstEnumeration() {
        return this.postorderEnumeration();
    }
    
    public Enumeration pathFromAncestorEnumeration(TreeNode treeNode) {
        return new DefaultMutableTreeNode.PathBetweenNodesEnumeration(treeNode, this);
    }
    
    public boolean isNodeChild(TreeNode treeNode) {
        return treeNode != null && this.getChildCount() != 0 && treeNode.getParent() == this;
    }
    
    public TreeNode getFirstChild() {
        if (this.getChildCount() == 0) {
            throw new NoSuchElementException("node has no children");
        }
        return this.getChildAt(0);
    }
    
    public TreeNode getLastChild() {
        if (this.getChildCount() == 0) {
            throw new NoSuchElementException("node has no children");
        }
        return this.getChildAt(this.getChildCount() - 1);
    }
    
    public TreeNode getChildAfter(TreeNode treeNode) {
        if (treeNode == null) {
            throw new IllegalArgumentException("argument is null");
        }
        int index = this.getIndex(treeNode);
        if (index == -1) {
            throw new IllegalArgumentException("node is not a child");
        }
        if (index < this.getChildCount() - 1) {
            return this.getChildAt(index + 1);
        }
        return null;
    }
    
    public TreeNode getChildBefore(TreeNode treeNode) {
        if (treeNode == null) {
            throw new IllegalArgumentException("argument is null");
        }
        int index = this.getIndex(treeNode);
        if (index == -1) {
            throw new IllegalArgumentException("argument is not a child");
        }
        if (index > 0) {
            return this.getChildAt(index - 1);
        }
        return null;
    }
    
    public boolean isNodeSibling(TreeNode treeNode) {
        int n;
        if (treeNode == null) {
            n = false;
        }
        else if (treeNode == this) {
            n = true;
        }
        else {
            TreeNode parent = this.getParent();
            n = (parent != null && parent == treeNode.getParent());
            if (n && !((DefaultMutableTreeNode)this.getParent()).isNodeChild(treeNode)) {
                throw new Error("sibling has different parent");
            }
        }
        return n;
    }
    
    public int getSiblingCount() {
        TreeNode parent = this.getParent();
        if (parent == null) {
            return 1;
        }
        return parent.getChildCount();
    }
    
    public DefaultMutableTreeNode getNextSibling() {
        DefaultMutableTreeNode defaultMutableTreeNode = (DefaultMutableTreeNode)this.getParent();
        DefaultMutableTreeNode defaultMutableTreeNode2;
        if (defaultMutableTreeNode == null) {
            defaultMutableTreeNode2 = null;
        }
        else {
            defaultMutableTreeNode2 = (DefaultMutableTreeNode)defaultMutableTreeNode.getChildAfter(this);
        }
        if (defaultMutableTreeNode2 != null && !this.isNodeSibling(defaultMutableTreeNode2)) {
            throw new Error("child of parent is not a sibling");
        }
        return defaultMutableTreeNode2;
    }
    
    public DefaultMutableTreeNode getPreviousSibling() {
        DefaultMutableTreeNode defaultMutableTreeNode = (DefaultMutableTreeNode)this.getParent();
        DefaultMutableTreeNode defaultMutableTreeNode2;
        if (defaultMutableTreeNode == null) {
            defaultMutableTreeNode2 = null;
        }
        else {
            defaultMutableTreeNode2 = (DefaultMutableTreeNode)defaultMutableTreeNode.getChildBefore(this);
        }
        if (defaultMutableTreeNode2 != null && !this.isNodeSibling(defaultMutableTreeNode2)) {
            throw new Error("child of parent is not a sibling");
        }
        return defaultMutableTreeNode2;
    }
    
    @Override
    public boolean isLeaf() {
        return this.getChildCount() == 0;
    }
    
    public DefaultMutableTreeNode getFirstLeaf() {
        while (!this.isLeaf()) {
            this = (DefaultMutableTreeNode)this.getFirstChild();
        }
        return this;
    }
    
    public DefaultMutableTreeNode getLastLeaf() {
        while (!this.isLeaf()) {
            this = (DefaultMutableTreeNode)this.getLastChild();
        }
        return this;
    }
    
    public DefaultMutableTreeNode getNextLeaf() {
        DefaultMutableTreeNode defaultMutableTreeNode = (DefaultMutableTreeNode)this.getParent();
        if (defaultMutableTreeNode == null) {
            return null;
        }
        DefaultMutableTreeNode nextSibling = this.getNextSibling();
        if (nextSibling != null) {
            return nextSibling.getFirstLeaf();
        }
        return defaultMutableTreeNode.getNextLeaf();
    }
    
    public DefaultMutableTreeNode getPreviousLeaf() {
        DefaultMutableTreeNode defaultMutableTreeNode = (DefaultMutableTreeNode)this.getParent();
        if (defaultMutableTreeNode == null) {
            return null;
        }
        DefaultMutableTreeNode previousSibling = this.getPreviousSibling();
        if (previousSibling != null) {
            return previousSibling.getLastLeaf();
        }
        return defaultMutableTreeNode.getPreviousLeaf();
    }
    
    public int getLeafCount() {
        int n = 0;
        Enumeration<Object> breadthFirstEnumeration = this.breadthFirstEnumeration();
        while (breadthFirstEnumeration.hasMoreElements()) {
            if (!((TreeNode)breadthFirstEnumeration.nextElement()).isLeaf()) {
                continue;
            }
            ++n;
        }
        if (n < 1) {
            throw new Error("tree has zero leaves");
        }
        return n;
    }
    
    @Override
    public String toString() {
        if (this.userObject == null) {
            return "";
        }
        return this.userObject.toString();
    }
    
    public Object clone() {
        DefaultMutableTreeNode defaultMutableTreeNode;
        try {
            defaultMutableTreeNode = (DefaultMutableTreeNode)super.clone();
            defaultMutableTreeNode.children = null;
            defaultMutableTreeNode.parent = null;
        }
        catch (CloneNotSupportedException ex) {
            throw new Error(ex.toString());
        }
        return defaultMutableTreeNode;
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        Object[] obj;
        if (this.userObject != null && this.userObject instanceof Serializable) {
            obj = new Object[] { "userObject", this.userObject };
        }
        else {
            obj = new Object[0];
        }
        objectOutputStream.writeObject(obj);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        Object[] array = (Object[])(Object[])objectInputStream.readObject();
        if (array.length > 0 && array[0].equals("userObject")) {
            this.userObject = array[1];
        }
    }
    
    static {
        EMPTY_ENUMERATION = Collections.emptyEnumeration();
    }
    
    private final class PreorderEnumeration implements Enumeration<TreeNode>
    {
        private final Stack<Enumeration> stack;
        
        public PreorderEnumeration(TreeNode obj) {
            this.stack = new Stack();
            Vector vector = new Vector(1);
            vector.addElement(obj);
            this.stack.push(vector.elements());
        }
        
        @Override
        public boolean hasMoreElements() {
            return !this.stack.empty() && ((Enumeration)this.stack.peek()).hasMoreElements();
        }
        
        @Override
        public TreeNode nextElement() {
            Enumeration<Object> enumeration = (Enumeration)this.stack.peek();
            TreeNode treeNode = (TreeNode)enumeration.nextElement();
            Enumeration<Object> children = treeNode.children();
            if (!enumeration.hasMoreElements()) {
                this.stack.pop();
            }
            if (children.hasMoreElements()) {
                this.stack.push(children);
                return treeNode;
            }
            return treeNode;
        }
    }
    
    final class PostorderEnumeration implements Enumeration<TreeNode>
    {
        protected TreeNode root;
        protected Enumeration<TreeNode> children;
        protected Enumeration<TreeNode> subtree;
        
        public PostorderEnumeration(TreeNode root) {
            this.root = root;
            this.children = this.root.children();
            this.subtree = DefaultMutableTreeNode.EMPTY_ENUMERATION;
        }
        
        @Override
        public boolean hasMoreElements() {
            return this.root != null;
        }
        
        @Override
        public TreeNode nextElement() {
            TreeNode root;
            if (this.subtree.hasMoreElements()) {
                root = (TreeNode)this.subtree.nextElement();
            }
            else if (this.children.hasMoreElements()) {
                this.subtree = new DefaultMutableTreeNode.PostorderEnumeration((TreeNode)this.children.nextElement());
                root = (TreeNode)this.subtree.nextElement();
            }
            else {
                root = this.root;
                this.root = null;
            }
            return root;
        }
    }
    
    final class BreadthFirstEnumeration implements Enumeration<TreeNode>
    {
        protected DefaultMutableTreeNode.BreadthFirstEnumeration.Queue queue;
        
        public BreadthFirstEnumeration(TreeNode obj) {
            Vector vector = new Vector(1);
            vector.addElement(obj);
            (this.queue = new DefaultMutableTreeNode.BreadthFirstEnumeration.Queue()).enqueue(vector.elements());
        }
        
        @Override
        public boolean hasMoreElements() {
            return !this.queue.isEmpty() && ((Enumeration)this.queue.firstObject()).hasMoreElements();
        }
        
        @Override
        public TreeNode nextElement() {
            Enumeration<Object> enumeration = (Enumeration)this.queue.firstObject();
            TreeNode treeNode = (TreeNode)enumeration.nextElement();
            Enumeration<Object> children = treeNode.children();
            if (!enumeration.hasMoreElements()) {
                this.queue.dequeue();
            }
            if (children.hasMoreElements()) {
                this.queue.enqueue(children);
                return treeNode;
            }
            return treeNode;
        }
        
        final class Queue
        {
            DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode head;
            DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode tail;
            
            public void enqueue(Object o) {
                if (this.head == null) {
                    DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode qNode = new DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode(o);
                    this.tail = qNode;
                    this.head = qNode;
                }
                else {
                    this.tail.next = new DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode(o);
                    this.tail = this.tail.next;
                }
            }
            
            public Object dequeue() {
                if (this.head == null) {
                    throw new NoSuchElementException("No more elements");
                }
                Object object = this.head.object;
                DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode head = this.head;
                this.head = this.head.next;
                if (this.head == null) {
                    this.tail = null;
                    return object;
                }
                head.next = null;
                return object;
            }
            
            public Object firstObject() {
                if (this.head == null) {
                    throw new NoSuchElementException("No more elements");
                }
                return this.head.object;
            }
            
            public boolean isEmpty() {
                return this.head == null;
            }
            
            final class QNode
            {
                public Object object;
                public DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode next;
                
                public QNode(Object object, DefaultMutableTreeNode.BreadthFirstEnumeration.Queue.QNode next) {
                    this.object = object;
                    this.next = next;
                }
            }
        }
    }
    
    final class PathBetweenNodesEnumeration implements Enumeration<TreeNode>
    {
        protected Stack<TreeNode> stack;
        
        public PathBetweenNodesEnumeration(TreeNode obj, TreeNode parent) {
            if (obj == null || parent == null) {
                throw new IllegalArgumentException("argument is null");
            }
            (this.stack = new Stack()).push(parent);
            while (parent != obj) {
                parent = parent.getParent();
                if (parent == null && parent != obj) {
                    throw new IllegalArgumentException("node " + obj + " is not an ancestor of " + parent);
                }
                this.stack.push(parent);
            }
            return;
        }
        
        @Override
        public boolean hasMoreElements() {
            return this.stack.size() > 0;
        }
        
        @Override
        public TreeNode nextElement() {
            try {
                return (TreeNode)this.stack.pop();
            }
            catch (EmptyStackException ex) {
                throw new NoSuchElementException("No more elements");
            }
        }
    }
}
